/*---------------------------------------------------------
 * Copyright (C) Microsoft Corporation. All rights reserved.
 *--------------------------------------------------------*/
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const registry_1 = require("./registry");
const grammarReader_1 = require("./grammarReader");
const theme_1 = require("./theme");
const grammar_1 = require("./grammar");
let DEFAULT_OPTIONS = {
    getGrammarDefinition: (scopeName) => null,
    getFilePath: (scopeName) => null,
    getInjections: (scopeName) => null
};
/**
 * The registry that will hold all grammars.
 */
class Registry {
    constructor(locator = DEFAULT_OPTIONS) {
        this._locator = locator;
        this._syncRegistry = new registry_1.SyncRegistry(theme_1.Theme.createFromRawTheme(locator.theme));
        this.installationQueue = new Map();
    }
    /**
     * Change the theme. Once called, no previous `ruleStack` should be used anymore.
     */
    setTheme(theme) {
        this._syncRegistry.setTheme(theme_1.Theme.createFromRawTheme(theme));
    }
    /**
     * Returns a lookup array for color ids.
     */
    getColorMap() {
        return this._syncRegistry.getColorMap();
    }
    /**
     * Load the grammar for `scopeName` and all referenced included grammars asynchronously.
     * Please do not use language id 0.
     */
    loadGrammarWithEmbeddedLanguages(initialScopeName, initialLanguage, embeddedLanguages) {
        return this.loadGrammarWithConfiguration(initialScopeName, initialLanguage, { embeddedLanguages });
    }
    /**
     * Load the grammar for `scopeName` and all referenced included grammars asynchronously.
     * Please do not use language id 0.
     */
    loadGrammarWithConfiguration(initialScopeName, initialLanguage, configuration) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._loadGrammar(initialScopeName);
            return this.grammarForScopeName(initialScopeName, initialLanguage, configuration.embeddedLanguages, configuration.tokenTypes);
        });
    }
    /**
     * Load the grammar for `scopeName` and all referenced included grammars asynchronously.
     */
    loadGrammar(initialScopeName) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._loadGrammar(initialScopeName);
        });
    }
    _loadGrammar(initialScopeName, dependentScope = null) {
        return __awaiter(this, void 0, void 0, function* () {
            // already installed
            if (this._syncRegistry.lookup(initialScopeName)) {
                return this.grammarForScopeName(initialScopeName);
            }
            // installation in progress
            if (this.installationQueue.has(initialScopeName)) {
                return this.installationQueue.get(initialScopeName);
            }
            // start installation process
            const prom = new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                let grammarDefinition = yield this._locator.getGrammarDefinition(initialScopeName, dependentScope);
                if (!grammarDefinition) {
                    throw new Error(`A tmGrammar load was requested but registry host failed to provide grammar definition`);
                }
                if (typeof grammarDefinition.content !== 'string' ||
                    typeof grammarDefinition.format !== 'string' ||
                    (grammarDefinition.format !== 'json' && grammarDefinition.format !== 'plist')) {
                    throw new TypeError(`Grammar definition must be an object ({content: string, format: 'json' | 'plist'})`);
                }
                const rawGrammar = grammarDefinition.format === 'json'
                    ? grammarReader_1.parseJSONGrammar(grammarDefinition.content, 'c://fakepath/grammar.json')
                    : grammarReader_1.parsePLISTGrammar(grammarDefinition.content, 'c://fakepath/grammar.plist');
                let injections = (typeof this._locator.getInjections === 'function') && this._locator.getInjections(initialScopeName);
                rawGrammar.scopeName = initialScopeName;
                let deps = this._syncRegistry.addGrammar(rawGrammar, injections);
                yield Promise.all(deps.map((scopeNameD) => __awaiter(this, void 0, void 0, function* () {
                    try {
                        return this._loadGrammar(scopeNameD, initialScopeName);
                    }
                    catch (error) {
                        throw new Error(`While trying to load tmGrammar with scopeId: '${initialScopeName}', it's dependency (scopeId: ${scopeNameD}) loading errored: ${error.message}`);
                    }
                })));
                resolve(this.grammarForScopeName(initialScopeName));
            }));
            this.installationQueue.set(initialScopeName, prom);
            yield prom;
            this.installationQueue.delete(initialScopeName);
            return prom;
        });
    }
    /**
     * Get the grammar for `scopeName`. The grammar must first be created via `loadGrammar` or `loadGrammarFromPathSync`.
     */
    grammarForScopeName(scopeName, initialLanguage = 0, embeddedLanguages = null, tokenTypes = null) {
        return this._syncRegistry.grammarForScopeName(scopeName, initialLanguage, embeddedLanguages, tokenTypes);
    }
}
exports.Registry = Registry;
exports.INITIAL = grammar_1.StackElement.NULL;
//# sourceMappingURL=main.js.map